@@ -683,7 +683,7 @@ int Assemble(struct supertype *st, char *mddev,
 			    > devices[most_recent].i.events)
 				most_recent = devcnt;
 		}
-		if (content->array.level == -4)
+		if (content->array.level == LEVEL_MULTIPATH)
 			/* with multipath, the raid_disk from the superblock is meaningless */
 			i = devcnt;
 		else
@@ -776,8 +776,8 @@ int Assemble(struct supertype *st, char *mddev,
 		/* note: we ignore error flags in multipath arrays
 		 * as they don't make sense
 		 */
-		if (content->array.level != -4)
-			if (!(devices[j].i.disk.state & (1<<MD_DISK_SYNC))) {
+		if (content->array.level != LEVEL_MULTIPATH)
+			if (!(devices[j].i.disk.state & (1<<MD_DISK_ACTIVE))) {
 				if (!(devices[j].i.disk.state
 				      & (1<<MD_DISK_FAULTY)))
 					sparecnt++;
@@ -990,7 +990,7 @@ int Assemble(struct supertype *st, char *mddev,
 	if (content->reshape_active) {
 		int err = 0;
 		int *fdlist = malloc(sizeof(int)* bestcnt);
-		if (verbose)
+		if (verbose > 0)
 			fprintf(stderr, Name ":%s has an active reshape - checking "
 				"if critical section needs to be restored\n",
 				chosen_name);
@@ -1008,7 +1008,7 @@ int Assemble(struct supertype *st, char *mddev,
 				fdlist[i] = -1;
 		}
 		if (!err)
-			err = Grow_restart(st, content, fdlist, bestcnt, backup_file, verbose);
+			err = Grow_restart(st, content, fdlist, bestcnt, backup_file, verbose > 0);
 		while (i>0) {
 			i--;
 			if (fdlist[i]>=0) close(fdlist[i]);
